/* -*- c-basic-offset: 3 -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 * 
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

#include <string.h>
#include <map>

#define CONVERTRGB( red, green, blue, pixel ) { \
   unsigned int _value = red + (green << 8) + (blue << 16); \
   if (_value == 0) { \
      pixel = 0; \
   } else { \
      int _label = colormap[_value]; \
      if (_label == 0) { \
	 pixel = ++current_label; \
	 colormap[_value] = pixel; \
      } else { \
	 pixel = _label; \
      } \
   } \
}

/**
 * Skip comments.
 * A comment is started with a # and is ended with an end of line.
 * @param f the file.
 */
void SkipComment( FILE *f ) {
   int c;
   
   do {
      c = fgetc(f);
      if (c == '#') {
	 do {
	    c = fgetc(f);
	 } while (c != '\n');
      }
   } while ((c == '#') || (c == '\r') || (c == '\t') || (c == '\n') || (c == ' '));
   ungetc(c, f);
}

bool isPPM( FILE *fp ) {
   int a = fgetc(fp);
   int b = fgetc(fp);
   return a == 'P' && ( b == '1' || b == '2' || b == '3' || b == '4' || b == '5' || b == '6');
}

/**
 * Converts a PPM image (Portable Pixmap) or PGM (Portable GrayMap)
 * to a suitable Pandore image.
 *
 * @param nomfichier	the filename of the ppm file.
 * @param objout	the result Pandore image.
 */
bool readPPPM( FILE *f, LabelImage<unsigned int> &image ) {
   long x, y;
   char imagecode[2];
   unsigned char *data;
   int type;
   int nbpix, nlig, ncol, ndep = 0;

   std::map<unsigned int, unsigned int> colormap;
   unsigned int current_label = 0;
   
   // HEADER
   // Read the magic number.
   if (fscanf(f, "%c", &imagecode[0]) < 1 || fscanf(f, "%c", &imagecode[1]) < 1) {
      return false;
   }
   if (imagecode[0] != 'P') {
      return false;
   }
   if (imagecode[1] == '2') { // Gray ascii.
      type = 2;
   } else if (imagecode[1] == '5') { // RAW Gray scale
      type = 5;
   } else if (imagecode[1] == '3') { // ASCII Color.
      type = 3;
   } else if (imagecode[1] == '6') { // RAW Color.
      type = 6;
   } else if (imagecode[1] == '1') { // ASCII Binary
      type = 1;
   } else if (imagecode[1] == '4') { // RAW Binary
      type = 4;
   } else {
      return false;
   }

#ifndef NDEBUG
   printf("Type = %d\n", type);
#endif

   SkipComment(f);
   if (fscanf(f, "%d", &ncol) < 1) {
      return false;
   }
   if (fscanf(f, "%d", &nlig) < 1) {
      return false;
   }

   char buf[255];
   if (fgets(buf, 254, f) == NULL) {
      return false;
   }
   sscanf(buf, "%d", &ndep);
   SkipComment(f);

#ifndef NDEBUG
   printf("Image size = %d cols x %d rows x %d planes\n", ncol, nlig, ndep);
#endif

   if (type == 2) { // P2: Gray ascii
      if (fscanf(f, "%d", &nbpix) < 1) {
	 return false;
      }
      nbpix = (nbpix / 256) + 1;
      SkipComment(f);
#ifndef NDEBUG
      printf("Pixel size = %d\n", nbpix);
#endif

      if (ndep == 0) {
	 if (nbpix == 1) {
	    image.New(ncol, nlig);
	    for (y = 0; y < nlig; y++) {
	       for (x = 0 ; x < ncol; x++) {
		  unsigned int val;
		  if (fscanf(f, "%d", &val) < 1) {
		     return false;
		  }
		  if (val > current_label) {
		     current_label = val;
		  }
		  image(y, x) = val;
	       }
	    }
	    image.Labels(current_label);
	 } else {
	    image.New(ncol, nlig);
	    for (y = 0; y < nlig; y++) {
	       for (x = 0; x < ncol; x++) {
		  unsigned int val;
		  if (fscanf(f, "%d", &val) < 1) {
		     return false;
		  }
		  if (val > current_label) {
		     current_label = val;
		  }
		  image(y, x) = val;
	       }
	    }
	    image.Labels(current_label);
	 }
      }
   } else if (type == 3) { // P3: Color ascii
      if (fscanf(f, "%d", &nbpix) < 1) {
	 return false;
      }
      nbpix = (nbpix / 256) + 1;
      SkipComment(f);
#ifndef NDEBUG
      printf("Pixel size = %d\n", nbpix);
#endif

      if (nbpix == 1) {
	 image.New(ncol, nlig);
	 for (y = 0; y < nlig; y++) {
	    for (x = 0 ; x < ncol; x++) {
	       int red;
	       if (fscanf(f, "%d", &red) < 1) {
		  return false;
	       }
	       int green;
	       if (fscanf(f, "%d", &green) < 1) {
		  return false;
	       }
	       int blue;
	       if (fscanf(f, "%d", &blue) < 1) {
		  return false;
	       }
	       CONVERTRGB(red, green, blue, image(y, x));
	    }
	 }
	 image.Labels(current_label);
      } else {
	 image.New(ncol, nlig);
	 for (y = 0; y < nlig; y++) {
	    for (x = 0; x <  ncol; x++) {
	       int red;
	       if (fscanf(f, "%d", &red) < 1) {
		  return false;}
	       int green;
	       if (fscanf(f, "%d", &green) < 1) {
		  return false;
	       }
	       int blue;
	       if (fscanf(f, "%d", &blue) < 1) {
		  return false;
	       }
	       CONVERTRGB(red, green, blue, image(y, x));
	    }
	 }
	 image.Labels(current_label);
      }
   } else if (type == 5) { // P5: RAW Gray
      if (fscanf(f, "%d", &nbpix) < 1) {
	 return false;
      }
      nbpix = (nbpix / 256) + 1;
      SkipComment(f);
#ifndef NDEBUG
      printf("Pixel size = %d\n", nbpix);
#endif
      if (ndep == 0) { // 2D
	 if ((data = (unsigned char*)malloc(sizeof(unsigned char) * nbpix * nlig * ncol)) == NULL) {
	    return false;
	 }
	 if (fread((unsigned char*)data, sizeof(unsigned char), nbpix * ncol * nlig, f) < 1) {
	    return false;
	 }
	 if (nbpix == 1) {
	    unsigned char *p;
	    p = data;
	    for (y = 0; y < nlig; y++) {
	       for (x = 0; x < ncol; x++, p++) {
		  if (*p > current_label) {
		     current_label = *p;
		  }
		  image(y, x) = *p;
	       }
	    }
	    image.Labels(current_label);
	 } else {
	    unsigned short* p = (unsigned short *)data;
	    for (y = 0; y < nlig; y++) {
	       for (x = 0; x < ncol; x++, p++) {
		  if (*p > current_label) {
		     current_label = *p;
		  }
		  image(y, x) = (long)*p;
	       }
	    }
	    image.Labels(current_label);
	 }
	 free((char*)data);
      }
   } else if (type == 6) { // P6: RAW Color
      if (fscanf(f, "%d", &nbpix) < 1) {
	 return false;
      }
      nbpix = (nbpix / 256) +1;
      SkipComment(f);
#ifndef NDEBUG
      printf("Pixel size = %d\n", nbpix);
#endif
      if ((data = (unsigned char*)malloc(sizeof(unsigned char)*nbpix * 3 * nlig * ncol)) == NULL) {
	 return false;
      }
      if (fread((char*)data, sizeof(unsigned char), nbpix * 3 * ncol * nlig, f) < 1) {
	 return false;
      }

      if (nbpix == 1) {
	 unsigned char *p;
	 p = data;

	 for (y = 0; y < nlig; y++) {
	    for (x = 0; x < ncol; x++) {
	       int red = *(p++);
	       int green = *(p++);
	       int blue = *(p++);
	       CONVERTRGB(red, green, blue, image(y, x));
	    }
	 }
	 image.Labels(current_label);
      } else {
	 unsigned short * p = (unsigned short *)data;

	 for (y = 0; y < nlig; y++) {
	    for (x = 0 ; x < ncol; x++) {
	       int red = *(p++);
	       int green = *(p++);
	       int blue = *(p++);
	       CONVERTRGB(red, green, blue, image(y, x));
	    }
	 }
	 image.Labels(current_label);
      }      
      free((char*)data);
   } else if (type == 1) { // P2: ascii RAW
      SkipComment(f);
      for (y = 0; y < nlig; y++) {
	 for (x = 0; x < ncol; x++) {
	    int val;
	    if (fscanf(f, "%d", &val) < 1) {
	       return false;
	    }
	    unsigned int pixel = (!val) * 255;
	    if (pixel > current_label) {
	       current_label = pixel;
	    }
	    image(y, x) = pixel;
 	 }
      }
      image.Labels(current_label);
   } else if (type == 4) { // P4: ascii RAW
      SkipComment(f);
      for (y = 0; y < nlig; y++) {
	 for (x = 0; x < ncol; x += 8) {
	    char val;
	    if (fscanf(f, "%c", &val) < 1) {
	       return false;
	    }
	    for (int i = 0; i < 8; i++) {
	       if (x + (7 - i) >= ncol) {
		  break;
	       }
	       unsigned int pixel = 255 * (char)!((val >> i) & 0x01);
	       if (pixel > current_label) {
		  current_label = pixel;
	       }
	       image(y, x + (7 - i)) = pixel;
	    }
	 }
      }
      image.Labels(current_label);
   }
   return true;
}
